import { mkdir, writeFile } from "fs/promises";
import * as path from "path";
import sharp from "sharp";

// Supported image formats
const SUPPORTED_FORMATS = [
  ".png",
  ".jpg",
  ".jpeg",
  ".gif",
  ".bmp",
  ".webp",
  ".tiff",
];

// Orientation types
export enum Orientation {
  TOP = 1,
  LEFT = 2,
  BOTTOM = 3,
  RIGHT = 4,
}

// Interface for library options
export interface SpeechBubbleOptions {
  mirror?: boolean;
  orientation?: Orientation;
  speechBubblePath?: string;
}

export class ImageSpeechBubbleTransformer {
  private defaultSpeechBubblePath: string;

  constructor(assetsDir: string = path.join(__dirname, "assets")) {
    this.defaultSpeechBubblePath = path.join(assetsDir, "speech_bubble.png");
  }

  /**
   * Check if the file format is supported
   */
  private checkFileFormat(filePath: string): void {
    const ext = path.extname(filePath).toLowerCase();
    if (!SUPPORTED_FORMATS.includes(ext)) {
      throw new Error(
        `Unsupported file extension ${ext}. Only ${SUPPORTED_FORMATS.join(", ")} are supported.`
      );
    }
  }

  /**
   * Transform image with optional mirroring and rotation
   */
  private async transformImage(
    inputBuffer: Buffer,
    options: { mirror?: boolean; orientation?: Orientation } = {}
  ): Promise<Buffer> {
    let pipeline = sharp(inputBuffer);

    // Mirror (flip horizontally)
    if (options.mirror) {
      pipeline = pipeline.flop();
    }

    // Rotate based on orientation
    switch (options.orientation) {
      case Orientation.LEFT:
        pipeline = pipeline.rotate(270);
        break;
      case Orientation.BOTTOM:
        pipeline = pipeline.rotate(180);
        break;
      case Orientation.RIGHT:
        pipeline = pipeline.rotate(90);
        break;
      default:
        // No rotation for TOP or default
        break;
    }

    return pipeline.toBuffer();
  }

  /**
   * Process image with speech bubble
   */
  async processSpeechBubble(
    inputBuffer: Buffer,
    options: SpeechBubbleOptions = {}
  ): Promise<Buffer> {
    // Validate input
    if (!inputBuffer) {
      throw new Error("Input buffer is required");
    }

    // Use default speech bubble path if not provided
    const speechBubblePath =
      options.speechBubblePath || this.defaultSpeechBubblePath;

    // Check file formats
    this.checkFileFormat(speechBubblePath);

    // Get input image metadata
    const inputMetadata = await sharp(inputBuffer).metadata();

    // Read and resize speech bubble to match input image
    const speechBubbleBuffer = await sharp(speechBubblePath)
      .resize({
        width: inputMetadata.width,
        height: inputMetadata.height,
        fit: sharp.fit.cover,
        position: sharp.strategy.entropy,
      })
      .toBuffer();

    // Process main image
    const processedImage = await sharp(inputBuffer)
      .ensureAlpha() // Ensure alpha channel
      .composite([
        {
          input: await this.transformImage(speechBubbleBuffer, {
            mirror: options.mirror,
            orientation: options.orientation,
          }),
          blend: "xor",
        },
      ])
      .toBuffer();

    return processedImage;
  }

  /**
   * Convenience method to process and save image
   */
  async processAndSave(
    inputBuffer: Buffer,
    outputPath: string,
    options: SpeechBubbleOptions = {}
  ): Promise<void> {
    // Check output file format
    this.checkFileFormat(outputPath);

    // Process image
    const processedBuffer = await this.processSpeechBubble(
      inputBuffer,
      options
    );

    // Ensure output directory exists
    const outputDir = path.dirname(outputPath);
    await mkdir(outputDir, { recursive: true });

    // Save processed image
    await writeFile(outputPath, processedBuffer);
    console.log(`Image saved at ${path.resolve(outputPath)}`);
  }
}

// Example usage
export function createSpeechBubbleTransformer(assetsDir?: string) {
  return new ImageSpeechBubbleTransformer(assetsDir);
}
